# Copyright 1999-2014. Parallels IP Holdings GmbH. All Rights Reserved.
from collections import defaultdict
import os
import logging
import socket
import re
import itertools
from xml.dom import minidom
from xml.sax import make_parser, handler
import xml.etree.ElementTree as ET

try: import subprocess
except ImportError: import compat.subprocess as subprocess

import pmmcli_config
import migration_handler_config
from subproc import NonzeroExitException
from pmm_migr_remote_service import MigrationRemoteService
from pmm_api_xml_protocols import ScoutResult, MigrationObjectList, ObjTypeReseller, ObjTypeClient, ObjTypeDomain, \
domains, clients, IPMapping, IpPairType, TaskStatus, Working, WorkingProgress, Starting, Finished, MigrationObjectAdminInfo, MigrationObjectServer, \
MigrationObjectServerLicense, AgentCapabilityInfoType, AgentCapabilityComponentsType, AgentCapabilityResourceUsageType,\
AgentCapabilityMiscellaneousType, AgentCapabilityApacheModulesType, AgentCapabilityDatabaseServersType, AgentCapabilityNameValueType, AgentCapabilityDatabaseServerType, ObjAdminInfoType


_logger = logging.getLogger("pmm.migration_handler.remote_service")


class MigrationRemoteServiceUnix(MigrationRemoteService):
    def __init__(self, host, port, login, passwd, session):
        MigrationRemoteService.__init__(self, host, port, login, passwd, session)

    def addContext(self, cmd):
        cmd.var('PSA_PASSWORD', self.passwd)
        cmd.arg('--host=' + self.host)
        if self.port is not None:
            cmd.arg('--port=' + self.port)
        cmd.arg('--login=' + self.login)
        cmd.arg('--session-path=' + self.getSessionPath())

    def sendScout(self, system_type = None):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--send-scout')
        self.addContext(cmd)
        self.runCmdLine(cmd)

    def getScoutResult(self):
        scout_result_path = os.path.join(self.getSessionPath(), 'scout-result.xml')
        scout_result_file = open(scout_result_path, 'r')
        scout_rsult_str = scout_result_file.read()
        scout_result = ScoutResult.factory()
        scout_result.build(minidom.parseString(scout_rsult_str).childNodes[0])
        return scout_result

    def removeScout(self):
        pass

    def deployAgent(self, agent_name, agent_dir):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--deploy-agent')
        self.addContext(cmd)
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        self.runCmdLine(cmd)
        self.session.setFlag('agent_uploaded')

    def backupImpl(self, agent_name, agent_dir):
        self.session.setValue('step', 'dump')

        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--dump-selected')
        self.addContext(cmd)
        cmd.arg('--selected-objects-file=' + self.session.getSelectedObjectsPath())
        if os.path.isfile(self.session.getRemoteDbServersPath()):
            cmd.arg('--remote-db-servers-file=' + self.session.getRemoteDbServersPath())
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)

        selected_objects = self.session.getObject('selected_objects')
        if selected_objects is not None:
            migration_parameters = selected_objects.get_parameters()
            if migration_parameters is not None:
                content_transport = migration_parameters.get_content_transport()
                if content_transport == '':
                    content_transport = pmmcli_config.get().get_content_transport()
                if content_transport == 'rsync':
                    cmd.arg('--content-transport=rsync')
            if selected_objects.get_server() is not None:
                cmd.arg('--server')
                if selected_objects.get_server().get_license() is None:
                    cmd.arg('--no-license')
            if selected_objects.get_admin_info() is not None:
                cmd.arg('--admin-info')
        self.runCmdLine(cmd)

        #download dump
        self.session.setValue('step','download')
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        self.addContext(cmd)
        cmd.arg('--get-content')
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        dumps_dir = self.session.getValue('dst_temporary_directory')
        if dumps_dir is not None:
            cmd.arg('--dumps-dir=' + dumps_dir)
        transfer_log_location = os.path.join(self.session.getSessionPath(), 'transfer.result')
        self.session.setTransferStatus(TaskStatus.factory( working = Working( starting = Starting()) ) )
        try:
            self.runCmdLine(cmd)
            self.session.setTransferStatus(TaskStatus.factory( finished = Finished( status = 'success', log_location = transfer_log_location) ) )
        except NonzeroExitException, e:
            self.session.setTransferStatus(TaskStatus.factory( finished = Finished( status = 'error', log_location = transfer_log_location) ) )
            raise

    def getObjectsList(self, agent_name, agent_dir):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--dump-all')
        cmd.arg('--no-content')
        self.addContext(cmd)
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        self.runCmdLine(cmd)

        objects_list = self.makeObjectsListFromDump(agent_name)
        return objects_list

    def makeObjectsListFromDump(self, agent_name):
        dump_path = os.path.join(self.getSessionPath(), 'dump.xml')

        generator = ObjectsListGenerator()

        xml = ET.parse(dump_path)
        return generator.getObjectsList(xml.getroot(), agent_name)

    def removeData(self, agent_name, agent_dir):
        if agent_name is None or agent_dir is None:
            _logger.info("Agent is not uploaded to source host. Skip removing agent.")
            return

        self.session.setValue('step', 'remove-scout')
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--remove-agent')
        self.addContext(cmd)
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        self.runCmdLine(cmd)

    def selectAccounts(self, file_path):
        pass

    def getIpMap(self, file_path):
        dump_path = os.path.join(self.getSessionPath(), 'dump.xml')

        objects_list_raw_path = os.path.join(self.getSessionPath(), 'objects_list_raw')
        objects_list_raw = MigrationObjectList.factory()
        objects_list_raw.build(minidom.parse(objects_list_raw_path).childNodes[0])

        server_info = objects_list_raw.get_server()
        if server_info is not None:
            misc_settings = server_info.get_capability_info().get_miscellaneous().get_setting()
            ip_ftps = dict([(s.get_name()[5:], s.get_value() == 'true') for s in misc_settings if s.get_name().startswith('ftps_')])
        else:
            ip_ftps = {}

        selected_objects = load_selected_objects(open(file_path))
        generator = IpMappingGenerator()
        xml = ET.parse(dump_path)
        return generator.getIpMapping(xml.getroot(), selected_objects, ip_ftps)

    def clearStatus(self, agent_name, agent_dir):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--remove-dumping-status')
        self.addContext(cmd)
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        self.runCmdLine(cmd)

    def is_valid_ipv6_address(self, address):
        try:
            addr= socket.inet_pton(socket.AF_INET6, address)
        except socket.error:
            return False
        return True

    def checkMigrationParameters(self, address, migration_parameters, scout_result):
        # check for content-transport="rsync"
        content_transport = migration_parameters.get_content_transport()
        if content_transport == '':
            content_transport = pmmcli_config.get().get_content_transport()
        if content_transport == 'rsync':
            no_src_rsync = True
            os_info_object = scout_result.get_os_info()
            packages_installed = os_info_object.get_package_installed()
            for package_installed in packages_installed:
                if package_installed.get_name() == 'rsync':
                    no_src_rsync = False
                    break
            if no_src_rsync:
                return 211, 'Rsync is not available on the source host'
            output = subprocess.Popen(["rsync", "--version"], stdout=subprocess.PIPE).communicate()[0]
            s = re.search("^rsync\\s+version\\s+(([0-9]{1,2})\\.([0-9]{1,2})\\.([0-9]{1,2})).*$",output, re.DOTALL)
            if s is None:
                return 211, 'Rsync is not available on the destination host'
            if self.is_valid_ipv6_address(address):
                if int(s.group(2))<2 or ( int(s.group(2))==2 and ( int(s.group(3))<6 or ( int(s.group(3))==6 and int(s.group(4))<9) ) ):
                    return 211, "Your rsync version (%s) does not support IPv6 addresses. At least 2.6.9 version is required" % s.group(1)
        return 0, None

    def getStatusImpl(self, agent_name, agent_dir):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--get-dumping-status')
        self.addContext(cmd)
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)

        status = TaskStatus.factory()
        working = Working.factory()
        dumping = WorkingProgress.factory()
        proc = self.runCmdLine(cmd)
        status_output = cmd.get_stdout()
        _logger.debug("Got status: %s" % status_output)
        dumping.build(minidom.parseString(status_output).childNodes[0])
        working.dumping = dumping
        status.working = working

        return status

    def getDownloadSpeed(self):
        output = None
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--get-download-speed')
        self.addContext(cmd)
        try:
            proc = self.runCmdLine(cmd)
            output = cmd.get_stdout()
            _logger.debug("Got output: %s" % output)
        except NonzeroExitException, e:
            _logger.debug("There is an error during retrieving download speed: %s" % unicode(e))
        return output

# Did you know that...
#   ... `if_not_none` is implementation of `bind` method for `maybe` monad.
def if_not_none(func, value):
    return func(value) if value is not None else None

class ObjectsListGenerator(object):
    def parse_ip_address(self, node):
        type = node.find('ip-type').text
        address = node.find('ip-address').text
        return (type, IpPairType(source=address))

    def parse_ips(self, nodes):
        ips = map(self.parse_ip_address, nodes)
        return IPMapping(
            shared=[ip for type, ip in ips if type == 'shared'],
            exclusive=[ip for type, ip in ips if type == 'exclusive']
        )

    def parse_ip_pool(self, node):
        return [self.parse_ips(node.findall('ip'))]

    def parse_reseller(self, node):
        name = node.attrib['name']
        contact = node.attrib.get('contact', '')
        ip_pool = if_not_none(self.parse_ip_pool, node.find('ip_pool'))
        client_list = map(self.parse_client, node.findall('clients/client'))
        domain_list = map(self.parse_domain, node.findall('domains/domain'))
        return ObjTypeReseller(
            name=name,
            contact=contact,
            ip=ip_pool,
            clients=clients(client_list) if client_list else None,
            domains=domains(domain_list) if domain_list else None,
        )

    def parse_domain(self, node):
        name = node.attrib['name']
        ipNodes = node.findall('properties/ip')
        ip = [self.parse_ips(ipNodes)] if ipNodes else None
        capability_info = if_not_none(self.parse_capability_info, node.find('capability-info'))
        return ObjTypeDomain(name=name, ip=ip, capability_info=capability_info)

    def parse_name_value(self, node):
        return AgentCapabilityNameValueType(node.find('name').text, node.find('value').text)

    def parse_database_server(self, node):
        return AgentCapabilityDatabaseServerType(
            typexx = node.find('type').text,
            host = node.find('host').text,
            port = node.find('port').text,
            version = node.find('version').text,
            local = node.find('local').text,
        )

    def parse_capability_info(self, node):
        return AgentCapabilityInfoType(
            components = AgentCapabilityComponentsType(
                map(self.parse_name_value, node.findall('components/component'))
            ),
            resource_usage = AgentCapabilityResourceUsageType(
                map(self.parse_name_value, node.findall('resource-usage/resource'))
            ),
            miscellaneous = AgentCapabilityMiscellaneousType(
                map(self.parse_name_value, node.findall('miscellaneous/setting'))
            ),
            apache_modules = AgentCapabilityApacheModulesType(
                map(self.parse_name_value, node.findall('apache-modules/module'))
            ),
            database_servers = AgentCapabilityDatabaseServersType(
                map(self.parse_database_server, node.findall('database-servers/database-server'))
            ),
        )

    def parse_client(self, node):
        name = node.attrib['name']
        contact = node.attrib.get('contact', '')
        ip_pool = if_not_none(self.parse_ip_pool, node.find('ip_pool'))
        domain_list = map(self.parse_domain, node.findall('domains/domain'))
        return ObjTypeClient(
            name=name,
            contact=contact,
            ip=ip_pool,
            domains=domains(domain_list) if domain_list else None,
            vendor_login = node.attrib.get('vendor-login', '')
        )

    def parse_admin(self, node):
        return (
            map(self.parse_reseller, node.findall('resellers/reseller')),
            map(self.parse_client, node.findall('clients/client')),
            map(self.parse_domain, node.findall('domains/domain')),
            ObjAdminInfoType(),
        )

    def parse_server(self, node):
        return MigrationObjectServer(
            capability_info=self.parse_capability_info(node.find('capability-info'))
        )

    def parse_dump(self, node):
        resellers, clients, domains, admin_info = self.parse_admin(node.find('admin'))
        server = if_not_none(self.parse_server, node.find('server'))
        return MigrationObjectList(reseller=resellers, client=clients, domain=domains, admin_info=admin_info, server=server)

    def getObjectsList(self, node, agent_name):
        objects_list = self.parse_dump(node)
        if agent_name == 'PleskX':
            objects_list.set_admin_info(MigrationObjectAdminInfo())
            if (objects_list.get_server() is None):
                server = MigrationObjectServer()
                objects_list.set_server(server)
            objects_list.get_server().set_license(MigrationObjectServerLicense())
        return objects_list

def load_selected_objects(file):
    doc = minidom.parse(file)
    root = doc.documentElement
    obj_list = MigrationObjectList.factory()
    obj_list.build(root)

    def flatten(l):
        return sum(l, [])

    def flat_map(func, l):
        return flatten(map(func, l))

    def reseller_clients(res):
        clients = res.get_clients()
        return clients.get_client() if clients is not None else []

    def client_domains(client):
        domains = client.get_domains()
        return domains.get_domain() if domains is not None else []

    resellers = obj_list.get_reseller()
    clients = flat_map(reseller_clients, resellers) + obj_list.get_client()
    domains = flat_map(client_domains, clients) + obj_list.get_domain()

    reseller_names = [res.get_name() for res in resellers]
    client_names = [client.get_name() for client in clients]
    domain_names = [domain.get_name() for domain in domains if domain.get_exists_on_target() != 'true']

    return reseller_names, client_names, domain_names

class IpMappingGenerator(object):
    def getIpMapping(self, node, selected_objects, ip_ftps):
        ips = defaultdict(dict)
        for ip_node in self.select_ip_nodes(node, selected_objects):
            ip_type = ip_node.find('ip-type').text
            ip_address = ip_node.find('ip-address').text
            if ip_address not in ips[ip_type]:
                ips[ip_type][ip_address] = IpPairType(
                    source=ip_address,
                    source_ftps={None: None, True: 'true', False: 'false'}[ip_ftps.get(ip_address)],
                )

        return IPMapping(
            shared=ips['shared'].values(),
            exclusive=ips['exclusive'].values(),
        )

    def select_ip_nodes(self, root, selected_objects):
        reseller_paths = ['admin/resellers/reseller']

        client_paths = [
            'admin/clients/client',
            'admin/resellers/reseller/clients/client',
        ]

        domain_paths = [
            'admin/domains/domain',
            'admin/clients/client/domains/domain',
            'admin/resellers/reseller/clients/client/domains/domain',
        ]

        resellers, clients, domains = selected_objects

        for node in itertools.chain(
            self.filter_nodes_by_name(root, reseller_paths, resellers),
            self.filter_nodes_by_name(root, client_paths, clients),
        ):
            for ip_node in node.findall('ip_pool/ip'):
                yield ip_node

        for node in self.filter_nodes_by_name(root, domain_paths, domains):
            for ip_node in node.findall('properties/ip'):
                yield ip_node

    # Select all nodes using given `paths` whose `name` attribute
    # is present in `names`.
    def filter_nodes_by_name(self, node, paths, names):
        for n in itertools.chain(*map(node.findall, paths)):
            if n.attrib['name'] in names:
                yield n

